"""
 Conversion between coordinates systems
"""
import numpy as np
from navipy.maths.homogeneous_transformations import rotation_matrix
from navipy.scene import is_numeric_array
from navipy.maths.homogeneous_transformations\
    import compose_matrix


def cartesian_to_spherical(x, y, z):
    radius = np.sqrt(x**2 + y**2 + z**2)
    elevation = np.arctan2(z, np.sqrt(x**2 + y**2))
    azimuth = np.arctan2(y, x)
    spherical = np.zeros_like(x)
    spherical = np.tile(spherical[..., np.newaxis], (3,))
    return elevation, azimuth, radius


def spherical_to_cartesian(elevation, azimuth, radius=1):
    cartesian = np.zeros_like(elevation)
    cartesian = np.tile(cartesian[..., np.newaxis], (3,))
    x = np.cos(elevation) * np.cos(azimuth)
    y = np.cos(elevation) * np.sin(azimuth)
    z = np.sin(elevation)
    cartesian = radius * cartesian
    return x, y, z


def cartesian_to_spherical_vectors(opticFlow, angles, viewing_direction):
    """Now we need the optic Flow in the spherical coordinates.
    A vector in cartesian coordinates can be transform as one in
    the spherical coordinates following the transformation:

    A_rho    =+A_x.*cos(epsilon).*cos(phi)
              +A_y.*cos(epsilon).*sin(phi)
              +A_z.*sin(epsilon)
    A_epsilon=-A_x.*sin(epsilon).*cos(phi)
              -A_y.*sin(epsilon).*sin(phi)
              +A_z.*cos(epsilon)
    A_phi    =-A_x.*sin(phi)+A_y.*cos(phi)

    for epsilon in [-pi/2 +pi/2] and phi in [0 2pi]
    reverse tajectory, needed because the frame x,y,z is expressed in
    the orientation Yaw=pitch=roll=0"""
    if opticFlow is None:
        raise ValueError("opticFlow must not be None")
    if angles is None:
        raise ValueError("angles must not be None")
    if viewing_direction is None:
        raise ValueError("viewing direction must not be None")
    if (not isinstance(opticFlow, np.ndarray)):
        raise TypeError("vector must be of type np.ndarray")
    if opticFlow.shape[0] != 3:
        raise Exception("first dimension of optic flow vector\
                         must have size three")
    if not is_numeric_array(opticFlow):
        raise TypeError("opticFlow must be of numerical type")
    if (not isinstance(viewing_direction, list)) and\
       (not isinstance(viewing_direction, np.ndarray)):
        raise TypeError("angles must be list or np.ndarray")
    if not is_numeric_array(viewing_direction):
        raise TypeError("viewing_direction must be of numerical type")
    if len(viewing_direction) != 2:
        raise Exception("first dimension of viewing\
                         direction must be of size two")

    vec = opticFlow
    ypr=angles
    M = compose_matrix(scale=None, shear=None, angles=ypr, translate=None,
                       perspective=None, axes='rzyx')[:3, :3]
    vec = np.dot(np.transpose(M), vec)
    opticFlow = vec

    OFT_x = opticFlow[0]
    OFT_y = opticFlow[1]
    OFT_z = opticFlow[2]

    epsilon = viewing_direction[1]
    phi = viewing_direction[0]

    # print("OFTs", opticFlow)
    # print("ep,phi", epsilon, phi)

    rofterm1 = +OFT_x*np.cos(epsilon)*np.cos(phi)
    rofterm2 = +OFT_y*np.cos(epsilon)*np.sin(phi)
    rofterm3 = +OFT_z*np.sin(epsilon)
    rof = rofterm1+rofterm2+rofterm3

    vofterm1 = -OFT_x*np.sin(epsilon)*np.cos(phi)
    vofterm2 = -OFT_y*np.sin(epsilon)*np.sin(phi)
    vofterm3 = OFT_z*np.cos(epsilon)
    vof = vofterm1+vofterm2+vofterm3
    hof = -OFT_x*np.sin(phi) + OFT_y*np.cos(phi)

    return [rof, hof, vof]
