"""

"""
import numpy as np
import navipy.maths.constants as constants
import navipy.maths.quaternion as quat


def matrix(ai, aj, ak, axes='sxyz'):
    """Return homogeneous rotation matrix from Euler angles and axis sequence.

    ai, aj, ak : Euler's roll, pitch and yaw angles
    axes : One of 24 axis sequences as string or encoded tuple
    """
    try:
        firstaxis, parity, repetition, frame = constants._AXES2TUPLE[axes]
    except (AttributeError, KeyError):
        constants._TUPLE2AXES[axes]  # validation
        firstaxis, parity, repetition, frame = axes

    i = firstaxis
    j = constants._NEXT_AXIS[i + parity]
    k = constants._NEXT_AXIS[i - parity + 1]

    if frame:
        ai, ak = ak, ai
    if parity:
        ai, aj, ak = -ai, -aj, -ak

    si, sj, sk = np.sin(ai), np.sin(aj), np.sin(ak)
    ci, cj, ck = np.cos(ai), np.cos(aj), np.cos(ak)
    cc, cs = ci * ck, ci * sk
    sc, ss = si * ck, si * sk

    M = np.identity(4)
    if repetition:
        M[i, i] = cj
        M[i, j] = sj * si
        M[i, k] = sj * ci
        M[j, i] = sj * sk
        M[j, j] = -cj * ss + cc
        M[j, k] = -cj * cs - sc
        M[k, i] = -sj * ck
        M[k, j] = cj * sc + cs
        M[k, k] = cj * cc - ss
    else:
        M[i, i] = cj * ck
        M[i, j] = sj * sc - cs
        M[i, k] = sj * cc + ss
        M[j, i] = cj * sk
        M[j, j] = sj * ss + cc
        M[j, k] = sj * cs - sc
        M[k, i] = -sj
        M[k, j] = cj * si
        M[k, k] = cj * ci
    return M


def from_matrix(matrix, axes='sxyz'):
    """Return Euler angles from rotation matrix for specified axis sequence.

    axes : One of 24 axis sequences as string or encoded tuple

    Note that many Euler angle triplets can describe one matrix.
    """
    try:
        firstaxis, parity, \
            repetition, frame = constants._AXES2TUPLE[axes.lower(
            )]
    except (AttributeError, KeyError):
        constants._TUPLE2AXES[axes]  # validation
        firstaxis, parity, repetition, frame = axes

    i = firstaxis
    j = constants._NEXT_AXIS[i + parity]
    k = constants._NEXT_AXIS[i - parity + 1]

    M = np.array(matrix, dtype=np.float64, copy=False)[:3, :3]
    if repetition:
        sy = np.sqrt(M[i, j] * M[i, j] + M[i, k] * M[i, k])
        if sy > constants._EPS:
            ax = np.arctan2(M[i, j],  M[i, k])
            ay = np.arctan2(sy,	   M[i, i])
            az = np.arctan2(M[j, i], -M[k, i])
        else:
            ax = np.arctan2(-M[j, k],  M[j, j])
            ay = np.arctan2(sy,	   M[i, i])
            az = 0.0
    else:
        cy = np.sqrt(M[i, i] * M[i, i] + M[j, i] * M[j, i])
        if cy > constants._EPS:
            ax = np.arctan2(M[k, j],  M[k, k])
            ay = np.arctan2(-M[k, i],  cy)
            az = np.arctan2(M[j, i],  M[i, i])
        else:
            ax = np.arctan2(-M[j, k],  M[j, j])
            ay = np.arctan2(-M[k, i],  cy)
            az = 0.0

    if parity:
        ax, ay, az = -ax, -ay, -az
    if frame:
        ax, az = az, ax
    return ax, ay, az


def from_quaternion(quaternion, axes='sxyz'):
    """Return Euler angles from quaternion for specified axis sequence.
    """
    return from_matrix(quat.matrix(quaternion), axes)


def angle_rate_matrix(ai, aj, ak, axes='sxyz'):
    """
    Return the Euler Angle Rates Matrix

    from Diebels Representing Attitude: Euler Angles,
    Unit Quaternions, and Rotation, 2006
    """
    try:
        firstaxis, parity, repetition, frame = constants._AXES2TUPLE[axes]
    except (AttributeError, KeyError):
        constants._TUPLE2AXES[axes]  # validation
        firstaxis, parity, repetition, frame = axes
    i = firstaxis
    j = constants._NEXT_AXIS[i + parity]
    k = constants._NEXT_AXIS[i - parity + 1]
    veci = np.zeros(3)
    veci[i] = 1
    vecj = np.zeros(3)
    vecj[j] = 1
    veck = np.zeros(3)
    veck[k] = 1
    if frame == 0:  # static
        roti = matrix(ai, 0, 0, axes=axes)[:3, :3]
        rotj = matrix(0, aj, 0, axes=axes)[:3, :3]
        mat = np.vstack([veci, roti.dot(vecj), roti.dot(rotj).dot(veck)])
    else:
        rotk = matrix(0, 0, ak, axes=axes)[:3, :3].transpose()
        rotj = matrix(0, aj, 0, axes=axes)[:3, :3].transpose()
        mat = np.vstack([rotk.dot(rotj.dot(veci)), rotk.dot(vecj), veck])
    return mat


def angular_velocity(ai, aj, ak, dai, daj, dak,  axes='sxyz'):
    """
    Return the angular velocity
    """
    return angle_rate_matrix(ai, aj, ak, axes=axes).dot(np.vstack([dai,
                                                                   daj,
                                                                   dak]))
