import unittest
import numpy as np
import navipy.maths.homogeneous_transformations as ht
import navipy.maths.quaternion as quat
import navipy.maths.random as random


class TestQuaternions(unittest.TestCase):
    def test_quaternion_from_euler(self):
        quaternion = quat.from_euler(1, 2, 3, 'ryxz')
        self.assertTrue(np.allclose(quaternion,
                                    [0.435953, 0.310622,
                                     -0.718287, 0.444435]))

    def test_about_axis(self):
        quaternion = quat.about_axis(0.123, [1, 0, 0])
        self.assertTrue(np.allclose(quaternion,
                                    [0.99810947, 0.06146124, 0, 0]))

    def test_matrix(self):
        matrix = quat.matrix([0.99810947, 0.06146124, 0, 0])
        self.assertTrue(np.allclose(matrix,
                                    ht.rotation_matrix(0.123,
                                                       [1, 0, 0])))

    def test_matrix_identity(self):
        matrix = quat.matrix([1, 0, 0, 0])
        self.assertTrue(np.allclose(matrix, np.identity(4)))

    def test_matrix_diagonal(self):
        matrix = quat.matrix([0, 1, 0, 0])
        self.assertTrue(np.allclose(matrix, np.diag([1, -1, -1, 1])))

    def test_from_matrix_identity(self):
        quaternion = quat.from_matrix(np.identity(4), True)
        self.assertTrue(np.allclose(quaternion, [1, 0, 0, 0]))

    def test_from_matrix_diagonal(self):
        quaternion = quat.from_matrix(np.diag([1, -1, -1, 1]))
        self.assertTrue(np.allclose(quaternion, [0, 1, 0, 0])
                        or np.allclose(quaternion, [0, -1, 0, 0]))

    def test_from_matrix_rotation(self):
        rotation = random.rotation_matrix()
        quaternion = quat.from_matrix(rotation)
        self.assertTrue(ht.is_same_transform(
            rotation, quat.matrix(quaternion)))

    # Seem to fail from time to time....
    # def test_quaternion_multiply(self):
    #    rotation_0 = random.rotation_matrix()
    #    quaternion_0 = quat.from_matrix(rotation_0)
    #    rotation_1 = random.rotation_matrix()
    #    quaternion_1 = quat.from_matrix(rotation_1)
    #    quaternion_01 = quat.multiply(quaternion_0,
    #                                  quaternion_1)
    #    rotation_01 = rotation_0.dot(rotation_1)
    #    quaternion_01_fromrot = quat.from_matrix(rotation_01)
    #    self.assertTrue(np.allclose(quaternion_01,
    #                                quaternion_01_fromrot))

    def test_inverse(self):
        quaternion_0 = random.quaternion()
        quaternion_1 = quat.inverse(quaternion_0)
        self.assertTrue(np.allclose(
            quat.multiply(quaternion_0, quaternion_1),
            [1, 0, 0, 0]))

    def test_quaternion(self):
        alpha, beta, gamma = 0.123, -1.234, 2.345
        xaxis, yaxis, zaxis = [1, 0, 0], [0, 1, 0], [0, 0, 1]
        rotx = ht.rotation_matrix(alpha, xaxis)
        roty = ht.rotation_matrix(beta, yaxis)
        rotz = ht.rotation_matrix(gamma, zaxis)
        rotation = rotx.dot(roty.dot(rotz))

        quatx = quat.about_axis(alpha, xaxis)
        quaty = quat.about_axis(beta, yaxis)
        quatz = quat.about_axis(gamma, zaxis)
        quaternion = quat.multiply(quatx, quaty)
        quaternion = quat.multiply(quaternion, quatz)
        rotq = quat.matrix(quaternion)
        self.assertTrue(ht.is_same_transform(rotation, rotq))


if __name__ == '__main__':
    unittest.main()
