"""
Motion code
"""
from navipy.scene import check_scene
from navipy.scene import __spherical_indeces__
from navipy.scene import is_numeric_array
from navipy.maths.homogeneous_transformations\
    import compose_matrix
from navipy.maths.coordinates\
    import spherical_to_cartesian, cartesian_to_spherical_vectors
import numpy as np
import pandas as pd


def spherical_to_Vector(sp):
    """ translate 2D azimuth and elevation vector to 3D vector
        :param sp: np array to be translated from spherical to euclidean
                  coordinates
        :returns vector
        :rtype np.ndarray """
    if sp is None:
        raise ValueError("sp must not be None")
    if not isinstance(sp, np.ndarray):
        raise TypeError("vector must be list or array")
    if len(sp) != 2:
        raise ValueError("vector must contain two elements")
    if (not isinstance(sp[0], float)) and (not isinstance(sp[0], int)):
        raise ValueError("azimuth must be float or integer")
    if (not isinstance(sp[1], float)) and (not isinstance(sp[1], int)):
        raise ValueError("elevation must be float or integer")
    vector = np.array([np.dot(np.cos(sp[1]), np.cos(sp[0])),
                       np.dot(np.cos(sp[1]), np.sin(sp[0])),
                       np.sin(sp[1])])

    return vector


def OFSubroutineSpToVector(sphericCoord):
    """
OFSUBROUTINESPTOVECTOR recieves a pair of spheric coordinates and
converts them into a 3dim vector.

   :param sphericCoord - a Sx2 matrix with spheric coordinates

   :return vector a Sx3 matrix with the transformed coordinates
   :rtype np.ndarray

NOTE: this is NOT the normal spheic coordinate system!
   spheric coordinates given as azimuth (Longitude) and
   zenith (Latitude) given in geographic coordinate system) in radian
   [pi/2 0]= [0 1 0] left, [3*pi/2 0]= [0 -1 0] right,
   [0 pi/2]= [0 0 1] top,  [0 -pi/2]= [0 0 -1] bottom


   #################################################################
   Copyright (C) 2009-2014 J.P. Lindemann, C. Strub

   This file is part of the ivtools.
   https://opensource.cit-ec.de/projects/ivtools

   the Optic-Flow toolbox is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

   the Optic-Flow toolbox is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
    ##################################################################
"""
    sp = sphericCoord
    if sp is None:
        raise ValueError("sp must not be None")
    if not isinstance(sp, list):
        raise TypeError("vector must be list or array")
    if not isinstance(sp, np.ndarray) and not isinstance(sp, list):
        raise TypeError("vector must be list or array")
    if len(sp) != 2:
        raise ValueError("vector must contain two elements")
    if (not isinstance(sp[0], float)) and (not isinstance(sp[0], int)):
        raise ValueError("azimuth must be float or integer")
    if (not isinstance(sp[1], float)) and (not isinstance(sp[1], int)):
        raise ValueError("elevation must be float or integer")
    """
    vector = np.zeros((3,))
    vector = [np.dot(np.cos(sp[1]), np.cos(sp[0])),
              np.dot(np.cos(sp[1]), np.sin(sp[0])),
              np.sin(sp[1])]
    """
    vector = np.array(spherical_to_cartesian(sp[1], sp[0]))

    return vector


def OFSubroutineRotate(vec, rotAx, alpha):
    """
    OFSUBROUTINEROTATE rotates the given vector through the rotAxis as
    UNIT-vector

    :param vec - Vx3 matrix with  vectors to rotate
    :param rotAx - the axis to rotate around
    :param alpha - the angle to rotate in radian

    :return rotVec - the rotated vectors in the input matrix
    :rtype np.ndarray



    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %	Copyright (C) 2009-2014 J.P. Lindemann, C. Strub
    %
    %   This file is part of the ivtools.
    %   https://opensource.cit-ec.de/projects/ivtools
    %
    %   the Optic-Flow toolbox is free software: you can redistribute it
    %    and/or modify it under the terms of the GNU General Public License
    %    as published by the Free Software Foundation, either version 3 of
    %    the License, or (at your option) any later version.
    %
    %   the Optic-Flow toolbox is distributed in the hope that it will be
    %    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    %    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    %    GNU General Public License for more details.
    %
    %    You should have received a copy of the GNU General Public License
    %    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    """
    if vec is None:
        raise ValueError("vector must not be None")
    if rotAx is None:
        raise ValueError("rotation axis must not be None")
    if not isinstance(vec, list) and (not isinstance(vec, np.ndarray)):
        raise TypeError("vector must be list or array")
    if not isinstance(rotAx, np.ndarray) and (not isinstance(rotAx, list)):
        raise TypeError("rotation Axis must be list or array")
    if len(vec) != 3:
        raise ValueError("vector must contain two elements")
    if (not isinstance(rotAx[0], float)) and (not isinstance(rotAx[0], int)):
        raise ValueError("rotation Axis must be float or integer")
    if (not isinstance(rotAx[1], float)) and (not isinstance(rotAx[1], int)):
        raise ValueError("rotation Axis must be float or integer")
    if (not isinstance(rotAx[2], float)) and (not isinstance(rotAx[2], int)):
        raise ValueError("rotation Axis must be float or integer")
    if (not isinstance(alpha, float)) and (not isinstance(alpha, int)):
        raise ValueError("rotation vector must be float or integer")
    if (not isinstance(vec, np.ndarray)) and (not isinstance(vec, list)):
        raise TypeError("vector must be of type np.ndarray or list")
    if isinstance(vec, np.ndarray):
        if vec.shape[0] != 3:
            raise Exception("second dimension of vector must have size three")
    if not is_numeric_array(vec):
        raise TypeError("vec must be of numerical type")
    # scaling the axis to unit vector
    rotAx = rotAx / np.linalg.norm(rotAx)

    # this is an rotation matrix:
    Rot = [[np.cos(alpha) + np.power(rotAx[0], 2) * (1 - np.cos(alpha)),
            np.dot(rotAx[0], rotAx[1]) *
            (1 - np.cos(alpha)) - rotAx[2] * np.sin(alpha),
            np.dot(rotAx[0], rotAx[2]) *
            (1-np.cos(alpha)) + rotAx[1] * np.sin(alpha)],
           [np.dot(rotAx[1], rotAx[0]) *
            (1-np.cos(alpha)) + rotAx[2] * np.sin(alpha),
            np.cos(alpha)+np.power(rotAx[1], 2) *
            (1-np.cos(alpha)), np.dot(rotAx[1],
            rotAx[2]) * (1-np.cos(alpha))-rotAx[0] * np.sin(alpha)],
           [np.dot(rotAx[2], rotAx[0]) *
            (1-np.cos(alpha))-rotAx[1] * np.sin(alpha),
            np.dot(rotAx[2], rotAx[1]) *
            (1-np.cos(alpha)) + rotAx[0] * np.sin(alpha),
            np.cos(alpha) + np.power(rotAx[2], 2) * (1-np.cos(alpha))]]

    rotVec = np.transpose(np.dot(Rot, np.transpose(vec)))
    return rotVec


def OFSubroutineYawPitchRoll(vec, ypr, inverse):
    """
    OFSUBROUTINEYAWPITCHROLL applies yaw pitch and roll to a vector
       inverse = true to inverse the YawPitchRoll

       :param vec - a 3dim vector to rotate
       :param ypr - [yaw pitch roll]
       :param inverse - boolean to dertermine whether the inverse rotations or
           the normal rotation should be computed

       :returns rotated vector
       :rtype np.ndarray



    #############################################################################
    Copyright (C) 2009-2014 J.P. Lindemann, C. Strub

       This file is part of the ivtools.
       https://opensource.cit-ec.de/projects/ivtools

       the Optic-Flow toolbox is free software: you can redistribute it
        and/or modify it under the terms of the GNU General Public License
        as published by the Free Software Foundation, either version 3 of
        the License, or (at your option) any later version.

       the Optic-Flow toolbox is distributed in the hope that it will
        be useful,but WITHOUT ANY WARRANTY; without even the implied
        warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
        See the GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
    """
    if vec is None:
        raise ValueError("vector must not be None")
    if ypr is None:
        raise ValueError("angles (ypr) must not be None")
    if (not isinstance(vec, np.ndarray)) and (not isinstance(vec, list)):
        raise TypeError("vector must be of type np.ndarray")
    if isinstance(vec, list):
        if len(vec) != 3:
            raise Exception("second dimension of vector must have size three")
    else:
        if vec.shape[0] != 3:
            raise Exception("second dimension of vector must have size three")
    if not is_numeric_array(vec):
        raise TypeError("vec must be of numerical type")
    if (not isinstance(ypr, np.ndarray)) and (not isinstance(ypr, list)):
        raise TypeError("angle vector must be of type np.ndarray or list")
    if isinstance(ypr, list):
        if len(ypr) != 3:
            raise Exception("first dimension of angle vector must have\
                            size three")
    else:
        if ypr.shape[0] != 3:
            raise Exception("first dimension of angle vector\
                            must have size three")
    if not is_numeric_array(ypr):
        raise TypeError("angle vector must be of numerical type")
    if not isinstance(inverse, bool):
        raise TypeError("inverse must be of type bool")

    # rotations around abselute fix-axis
    # if (inverse==true)
    #     vec=OFSubroutineRotate(vec,[1 0 0],-ypr(3));
    #     vec=OFSubroutineRotate(vec,[0 1 0],-ypr(2));
    #     vec=OFSubroutineRotate(vec,[0 0 1],-ypr(1));
    # else
    #     vec=OFSubroutineRotate(vec,[0 0 1],ypr(1));
    #     vec=OFSubroutineRotate(vec,[0 1 0],ypr(2));
    #     vec=OFSubroutineRotate(vec,[1 0 0],ypr(3));
    # end

    # rotations around relative rotatet axis
    M = compose_matrix(scale=None, shear=None, angles=ypr, translate=None,
                       perspective=None, axes='rzyx')[:3, :3]
    if (inverse):
        vec = np.dot(np.transpose(M), vec)
        """
        vec=tmpvec
        M1 = rotation_matrix(-ypr[2], [1, 0, 0])[:3, :3]
        vec = np.transpose(np.dot(M1, np.transpose(vec)))
        roty = np.transpose(np.dot(M1, np.transpose([0, 1, 0])))
        M2 = rotation_matrix(-ypr[1], roty)[:3, :3]
        vec = np.transpose(np.dot(M2, np.transpose(vec)))
        rotz = np.transpose(np.dot(M1, np.transpose([0, 0, 1])))
        M4 = rotation_matrix(-ypr[1], [0, 1, 0])[:3, :3]
        rotatedax = np.transpose(np.dot(M4, np.transpose(rotz)))

        M3 = rotation_matrix(-ypr[0], rotatedax)[:3,:3]
        vec = np.transpose(np.dot(M3, np.transpose(vec)))
        """
    else:
        vec = np.dot(M, vec)
        """
        print("new normal vec", vec)
        vec = tmpvec
        M1 = rotation_matrix(ypr[0], [0, 0, 1])[:3, :3]
        vec = np.transpose(np.dot(M1, np.transpose(vec)))
        roty = np.transpose(np.dot(M1, np.transpose([0, 1, 0])))
        M2 = rotation_matrix(ypr[1], roty)[:3, :3]
        vec = np.transpose(np.dot(M2, np.transpose(vec)))
        rotx = np.transpose(np.dot(M1, np.transpose([1, 0, 0])))
        M4 = rotation_matrix(ypr[1], [0, 1, 0])[:3, :3]
        rotatedax = np.transpose(np.dot(M4, np.transpose(rotx)))

        M3 = rotation_matrix(ypr[2], rotatedax)[:3, :3]
        vec = np.transpose(np.dot(M3, np.transpose(vec)))
        print("old normal vec", vec)
        """

    return vec


def optic_flow(scene, viewing_directions, velocity, distance_channel=3):
    """ optic flow
    :param scene: ibpc
    :param viewing_directions: viewing direction of each pixel
           (azimuth,elevation)
    :param velocity: pandas series
                     (x,y,z,alpha,beta,gamma,dx,dy,dz,dalpha,dbeta,dgamma)
    :distance channel: distance"""
    check_scene(scene)
    if distance_channel not in range(4):
        raise ValueError('distance channel out of range')
    if not isinstance(velocity, pd.Series):
        raise TypeError('velocity should be a pandas Series')
    if velocity is None:
        raise ValueError("velocity must not be None")
    if velocity.empty:
        raise Exception('velocity must not be empty')
    if not isinstance(velocity.index, pd.core.index.MultiIndex):
        raise Exception('velocity must have a multiindex containing \
                         the convention used')
    index = velocity.index
    convention = index.get_level_values(0)[-1]
    if convention != 'rxyz':
        raise ValueError("to calculate the optic flow,\
                          angles must be in rxyz convention")
    if 'x' not in velocity.index.get_level_values(1):
        raise ValueError('missing index x')
    if 'y' not in velocity.index.get_level_values(1):
        raise ValueError('missing index y')
    if 'z' not in velocity.index.get_level_values(1):
        raise ValueError('missing index z')
    if 'alpha_0' not in velocity.index.get_level_values(1):
        raise ValueError('missing index alpha_0')
    if 'alpha_1' not in velocity.index.get_level_values(1):
        raise ValueError('missing index alpha_1')
    if 'alpha_2' not in velocity.index.get_level_values(1):
        raise ValueError('missing index alpha_2')
    if np.any(pd.isnull(velocity)):
        raise ValueError('velocity must not contain nan')
    if viewing_directions is None:
        raise ValueError("viewing direction must not be None")
    if (not isinstance(viewing_directions, list)) and\
       (not isinstance(viewing_directions, np.ndarray)):
        raise TypeError("angels must be list or np.ndarray")
    if not is_numeric_array(viewing_directions):
        raise TypeError("viewing_direction must be of numerical type")
    if viewing_directions.shape[1] != 2:
        raise Exception("second dimension of viewing\
                         direction must be of size two")
    distance = np.squeeze(scene[:, :, distance_channel, 0])
    # distance += distance
    elevation = viewing_directions[..., __spherical_indeces__['elevation']]
    azimuth = viewing_directions[..., __spherical_indeces__['azimuth']]
    yaw = velocity[convention]['alpha_0']
    pitch = velocity[convention]['alpha_1']
    roll = velocity[convention]['alpha_2']
    dyaw = velocity[convention]['dalpha_0']
    dpitch = velocity[convention]['dalpha_1']
    droll = velocity[convention]['dalpha_2']

    if ((np.abs(dyaw) > np.pi/2 and 2*np.pi - np.abs(dyaw) > np.pi/2) or
       (np.abs(dpitch) > np.pi/2 and 2*np.pi - np.abs(dpitch) > np.pi/2) or
       (np.abs(droll) > np.pi/2 and 2*np.pi - np.abs(droll) > np.pi/2)):
        raise ValueError('rotation exceeds 90°, computation aborted')

    u = [velocity['location']['dx'],
         velocity['location']['dy'],
         velocity['location']['dz']]
    v = np.linalg.norm(u)
    if(v == 0):
        u = [0, 0, 0]
    else:
        u = u/np.linalg.norm(u)

    yawNow = OFSubroutineYawPitchRoll([1, 0, 0], [yaw, 0, 0], False)
    yawNext = OFSubroutineYawPitchRoll([1, 0, 0], [yaw+dyaw, 0, 0], False)
    rYaw = np.cross(yawNow, yawNext)

    # now rotation around y-axis
    pitchNow = OFSubroutineYawPitchRoll([1, 0, 0], [0, pitch, 0], False)
    pitchNext = OFSubroutineYawPitchRoll([1, 0, 0],
                                         [0, pitch+dpitch, 0], False)
    rPitch = np.cross(pitchNow, pitchNext)

    # and roatation around x-axis
    rollNow = OFSubroutineYawPitchRoll([0, 0, 1], [0, 0, roll], False)
    rollNext = OFSubroutineYawPitchRoll([0, 0, 1], [0, 0, roll+droll], False)
    rRoll = np.cross(rollNow, rollNext)

    rof = np.zeros((azimuth.shape[0], elevation.shape[0]))
    hof = np.zeros((azimuth.shape[0], elevation.shape[0]))
    vof = np.zeros((azimuth.shape[0], elevation.shape[0]))

    for i, a in enumerate(azimuth):
        for j, e in enumerate(elevation):
            spline = OFSubroutineSpToVector([a, e])
            spline = OFSubroutineYawPitchRoll(spline,
                                              [yaw, pitch, roll],
                                              False)
            p = distance[j, i]
            if(p == 0):
                # if object touches the enviroment, OpticFlow dosnt need to be
                # scaled -> distance=1
                p = 1

            # the Translation-part of the Optic Flow:
            dotvu = np.dot(u, v)
            splinetrans = np.transpose(spline)
            opticFlowT = -(dotvu-(np.dot(dotvu, splinetrans))*spline)/p
            # check if there actualy is a rotation and if compute
            # surface-normal and scale it with angle between vectors
            # negative because flow relative to observer
            if np.linalg.norm(rYaw) <= 0.000001:
                opticFlowRyaw = 0
            else:
                normrYaw = np.linalg.norm(rYaw)
                rYawN = np.dot(rYaw/normrYaw, np.arcsin(normrYaw))
                opticFlowRyaw = -np.cross(rYawN, spline)

            if np.linalg.norm(rPitch) <= 0.000001:
                opticFlowRpitch = 0
            else:
                normrPitch = np.linalg.norm(rPitch)
                rPitchN = np.dot(rPitch/normrPitch, np.arcsin(normrPitch))
                opticFlowRpitch = -np.cross(rPitchN, spline)
            if np.linalg.norm(rRoll) <= 0.000001:
                opticFlowRroll = 0
            else:
                normrRoll = np.linalg.norm(rRoll)
                rRollN = rRoll/normrRoll*np.arcsin(normrRoll)
                opticFlowRroll = -np.cross(rRollN, spline)

            # combine the rotations
            opticFlowR = opticFlowRyaw+opticFlowRpitch+opticFlowRroll

            # and add Translation and Rotation to get the Optic Flow
            opticFlow = opticFlowT+opticFlowR

            # Transform OF from Cartesian coordinates to Spherical coordinates
            # according to method
            (OF_rho, OF_phi, OF_epsilon) =\
                cartesian_to_spherical_vectors(opticFlow,
                                               [yaw, pitch, roll],
                                               [a, e])

            rof[j, i] = OF_rho
            hof[j, i] = OF_phi
            vof[j, i] = OF_epsilon

    return rof, hof, vof


class Module():
    """
    This class represents a Module that functions as a storage
    """
    def __init__(self, size=(180, 360)):
        """
        initializes the storage as an np.ndarray containing zeros
        of size size
        :param size: the tupel containing the size of the
                     storage (Input)
        """
        if size is None:
            raise ValueError("size must not be None")
        if not isinstance(size, tuple):
            raise TypeError("size must be a tuple")
        if len(size) < 2:
            raise Exception("length of size must at least be two")
        self.size = size
        self.Input = np.zeros(size)

    @property
    def size(self):
        """
        getter for the the size field
        :returns size: size of the Input field
        :rtype tuple
        """
        return self.__size

    @size.setter
    def size(self, size):
        """
        setter for the size of the storage
        :param size: tuple that contains the size of the storage
        """
        if size is None:
            raise ValueError("size must not be None")
        if not isinstance(size, tuple):
            raise TypeError("size must be a tuple")
        if len(size) < 2:
            raise Exception("length of size must at least be two")
        self.__size = size

    @property
    def Input(self):
        """
        getter for the Input field
        :returns Input
        :rtype np.ndarray
        """
        return self.__Input

    @Input.setter
    def Input(self, Input):
        """
        setter for the Input field, automaticaly sets the
        the size field to the shape of the Input
        :param Input
        """
        if Input is None:
            raise ValueError("Input must not be None")
        if not isinstance(Input, np.ndarray):
            raise TypeError("Input must be np array")
        if len(Input.shape) < 2:
            raise Exception("Input must have at least 2 dimensions")
        if (Input.shape[0] < 1) and (Input.shape[1] < 1):
            raise Exception("Each dimension of the Input\
                             must have at least be of size one")
        self.__Input = Input
        self.size = Input.shape

    def update(self,):
        """"
        update function can be implemented for
        inheriting classes
        """
        pass


class lp(Module):
    """
    Implementation of a low pass filter
    """
    def __init__(self, tau, inM):
        """
        Initializes the lowpass filter, the size of the output is
        set to the size of the input signal (inM)
        :param tau: time constant of the filter
        :param freq: cut off frequence of the filter
        :param inM: Module that stores and represents the input signal
        """
        if not isinstance(tau, float) and isinstance(tau, int):
            raise ValueError("tau must be of type float or integer")
        if inM is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM, Module):
            raise ValueError("input Module must be of type Module")
        self.size = inM.size
        Module.__init__(self, self.size)
        self.inM = inM
        self.itau = tau

    @property
    def itau(self):
        """
        getter of the time constant which is
        calculated by 1000/(tau*freq)
        """
        return self.__itau

    @itau.setter
    def itau(self, itau):
        """
        setter of the time constant
        :param itau: time constant
        """
        if not isinstance(itau, float) and isinstance(itau, int):
            raise ValueError("itau must be of type float or integer")
        self.__itau = itau

    @property
    def inM(self):
        """
        setter of the input Module
        :returns inM
        :rtype Module
        """
        return self.__inM

    @inM.setter
    def inM(self, inM):
        """
        setter of the input Module
        :param inM
        """
        if inM is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM = inM

    def update(self,):
        """
        update functions, updates the filtered signal for the
        the current input signal. out_t+1 +=tau*(input-out_t)
        """
        In = self.inM.Input
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                self.Input[i, j] += self.itau*(In[i, j]-self.Input[i, j])


class hp(Module):  # for second order just take hp for inM
    """
    Implements a high pass filter
    """
    def __init__(self, tau, inM):
        """
        Initializes the high pass filter
        :param tau: time constant
        :param freq: cut off frequency
        :param inM: Module that stores the input signal
        """
        if not isinstance(tau, float) and isinstance(tau, int):
            raise ValueError("tau must be of type float or integer")
        if inM is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM, Module):
            raise ValueError("input Module must be of type Module")
        self.size = inM.size
        Module.__init__(self, self.size)
        self.inM = inM
        self.lowpass = lp(tau, inM)

    @property
    def inM(self):
        """
        getter for the input Module
        :returns inM
        :rtype Module
        """
        return self.__inM

    @inM.setter
    def inM(self, inM):
        """
        setter for the input Module
        :param inM: input Module for the input signal
        """
        if inM is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM = inM

    def update(self,):
        """
        updates the output signal with the current input signal
        out_t+1=Input-lowpass(Input)
        """
        self.inM.update()
        self.lowpass.update()
        lpOut = self.lowpass.Input
        In = self.inM.Input
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                self.Input[i, j] = (In[i, j]-lpOut[i, j])


class mul(Module):
    """
    Implements the multiplication of two Modules
    """
    def __init__(self, inM1, inM2):
        """
        Initializes the multiplication module
        :param inM1: first input Module
        :param inM2: second input Module
        """
        if inM1 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM1, Module):
            raise ValueError("input Module must be of type Module")
        if inM2 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM2, Module):
            raise ValueError("input Module must be of type Module")
        Module.__init__(self, inM1.size)
        self.inM1 = inM1
        self.inM2 = inM2

    @property
    def inM1(self):
        """
        getter for the first input Module
        :returns inM1
        :rtype Module
        """
        return self.__inM1

    @inM1.setter
    def inM(self, inM1):
        """
        setter for the first input Module
        :param inM1
        """
        if inM1 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM1, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM1 = inM1

    @property
    def inM2(self):
        """
        getter for the second input Module
        :returns inM2
        :rtype Module
        """
        return self.__inM2

    @inM2.setter
    def inM2(self, inM2):
        """
        setter for the first input Module
        :param inM1
        """
        if inM2 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM2, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM2 = inM2

    def update(self, shift, axis=None):
        """
        updates the output (multiplication of the two input Modules)
        for the current input (see numpy roll)
        :param shift: shifts the Input provided by the first module
                      by the given amount
        :param axis: shifts the Input of the first module along the
                     provided axis
        """
        if not isinstance(shift, int):
            raise TypeError("shift must be an integer")
        if axis is not None:
            if not isinstance(axis, int):
                raise TypeError("axis must be of type integer")
        shiftedInput = np.roll(self.inM1.Input, shift, axis)
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                sig_left1 = self.inM1.Input[i, j]
                sig_left2 = shiftedInput[i, j]
                self.Input[i, j] = sig_left1*sig_left2


class div(Module):
    """
    Implements the division of two Modules
    """
    def __init__(self, inM1, inM2):
        """
        Initializes the multiplication module
        :param inM1: first input Module
        :param inM2: second input Module
        """
        if inM1 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM1, Module):
            raise ValueError("input Module must be of type Module")
        if inM2 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM2, Module):
            raise ValueError("input Module must be of type Module")
        Module.__init__(self, inM1.size)
        self.inM1 = inM1
        self.inM2 = inM2

    @property
    def inM1(self):
        """
        getter for the first input Module
        :returns inM1
        :rtype Module
        """
        return self.__inM1

    @inM1.setter
    def inM(self, inM1):
        """
        setter for the first input Module
        :param inM1
        """
        if inM1 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM1, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM1 = inM1

    @property
    def inM2(self):
        """
        getter for the second input Module
        :returns inM2
        :rtype Module
        """
        return self.__inM2

    @inM2.setter
    def inM2(self, inM2):
        """
        setter for the first input Module
        :param inM1
        """
        if inM2 is None:
            raise ValueError("input Module must not be None")
        if not isinstance(inM2, Module):
            raise ValueError("input Module must be of type Module")
        self.__inM2 = inM2

    def update(self, shift, axis=None):
        """
        updates the output (division of the two input Modules)
        for the current input (see numpy roll)
        :param shift: shifts the Input provided by the first module
                      by the given amount
        :param axis: shifts the Input of the first module along the
                     provided axis
        """
        if not isinstance(shift, int):
            raise TypeError("shift must be an integer")
        if axis is not None:
            if not isinstance(axis, int):
                raise TypeError("axis must be of type integer")
        shiftedInput = np.roll(self.inM1.Input, shift, axis)
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                sig_left1 = self.inM1.Input[i, j]
                sig_left2 = shiftedInput[i, j]
                if sig_left2 != 0:
                    self.Input[i, j] = sig_left1/sig_left2
                else:
                    self.Input[i, j] = 0
